# Would like to use auto file loader, but it likes to hide bugs!
from unstructured.partition.xlsx import partition_xlsx
from unstructured.partition.docx import partition_docx
from unstructured.partition.doc import partition_doc
from unstructured.partition.html import partition_html
from unstructured.partition.epub import partition_epub
from unstructured.partition.odt import partition_odt
from unstructured.partition.ppt import partition_ppt
from unstructured.partition.pptx import partition_pptx
from unstructured.partition.tsv import partition_tsv

import warnings
with warnings.catch_warnings():  # This rigamarol prevents these stupid SwigPy warnings caused by pymupdf
    warnings.simplefilter("ignore")
    import fitz


# An iterable of RawChunks is returned by load_and_split
class RawChunk():
    page_content: str
    metadata: dict
    def __init__(self, page_content, metadata={}):
        self.page_content = page_content
        self.metadata = metadata

"""

Takes string and returns list of strings

The algorithm:
- Split on the first separator
- Combine chunks in order if possible (i.e., resulting chunk isn't too big)
- Any chunks that are too big get the recursive split called on them again

Why a class instead of a function? Mostly just to match langchain behavior, or to specify args just once before passing same parametrized text splitter elsewhere, like in a retriever.

"""
class TextSplitter():
    # max_chunk_size is in units determined by length function; by default, characters.
    def __init__(self, max_chunk_size=2048, length_function=len, chunk_overlap=0):
        self.max_chunk_size = max_chunk_size
        self.length_function=length_function
        self.chunk_overlap = 0  # NOTE: doesn't yet support anything > 0 (in here for legacy langchain reasons)

    def split_text(self, txt, separators=None):

        DEFAULT_SEPARATORS = [  # the point of the weird character codes is to support multiple languages. Recommended here: https://python.langchain.com/v0.1/docs/modules/data_connection/document_transformers/recursive_text_splitter/
            "\n\n",
            "\n",
            "\t",
            " ",
            ".",
            ",",
            "\u200b",  # Zero-width space
            "\uff0c",  # Fullwidth comma
            "\u3001",  # Ideographic comma
            "\uff0e",  # Fullwidth full stop
            "\u3002",  # Ideographic full stop
        ]
        separators = separators if separators else DEFAULT_SEPARATORS

        def recursive_split(text, seps):
            max_chunks = []  # Holds chunks of their maximum size - not necessarily under max_chunk_size yet, though they'll never get bigger.

            sep = ""
            if not len(seps):
                # Means that some chunk is too big even after exhausting separators
                # Now just need to go in and dice it up without regard to separators
                splitsville = [text[:self.max_chunk_size], text[self.max_chunk_size:]]
            else:
                sep = seps[0]
                # Split on separator 
                splitsville = text.split(sep)

            # Max-combine chunks in order
            for split in splitsville:
                if not split:  # blanks are not allowed
                    continue
                if not len(max_chunks):  # first split creates new chunk 
                    max_chunks.append(split)
                elif self.length_function(max_chunks[len(max_chunks) - 1] + sep + split) < self.max_chunk_size:  # if we can add the current split to the previous without overflow, do it
                    max_chunks[len(max_chunks) - 1] += sep + split
                else:  # Otherwise, create a new chunk
                    max_chunks.append(split)

            # On any max_chunks that are not under max_chunk_size, split it again.
            final_chunks = []
            for max_chunk in max_chunks:
                if self.length_function(max_chunk) > self.max_chunk_size:
                    new_chunks = recursive_split(max_chunk, seps[1:])
                    final_chunks.extend(new_chunks)
                else:
                    final_chunks.append(max_chunk)
            
            return final_chunks
        
        if self.length_function(txt) <= self.max_chunk_size:
            if not txt:
                return []
            return [txt]
        
        return recursive_split(txt, separators)


class FileLoader():
    def __init__(self, fname):
        self.fname = fname

    # Returns iterable of RawChunks
    def load_and_split(self, text_splitter: TextSplitter):
        raise NotImplementedError("Function load_and_split not implemented for some file loader.")


# "Unstructured" being the name of a package
class UnstructuredLoader(FileLoader):
    def load_and_split(self, text_splitter):
        # This corresponds to mode=single in langchain
        elements = self._get_elements()
        total = "\n\n".join([str(el) for el in elements])
        splitsville = text_splitter.split_text(total)
        for split in splitsville:
            yield RawChunk(split, {})
    
    # For different file types, returns iterable
    def _get_elements(self):
        raise NotImplementedError("Function _get_elements not implemented for some file loader.")


class ExcelLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_xlsx(filename=self.fname)


class DocxLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_docx(filename=self.fname, infer_table_structure=True)


class HTMLLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_html(filename=self.fname)


class EpubLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_epub(filename=self.fname)


class OdtLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_odt(filename=self.fname)


class DocLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_doc(filename=self.fname)


class PptLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_ppt(filename=self.fname)


class PptxLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_pptx(filename=self.fname)


class TsvLoader(UnstructuredLoader):
    def _get_elements(self):
        return partition_tsv(filename=self.fname)


# Not using unstructured for CSV because it has a littany of bugs.
class CsvLoader(UnstructuredLoader):
    def load_and_split(self, text_splitter: TextSplitter):
        with open(self.fname, 'r') as file:
            original_contents = file.read()
        splitsville = text_splitter.split_text(original_contents)
        for split in splitsville:
            yield RawChunk(split, {})


# Divides by page, and then splits within each page.
class PdfLoader(FileLoader):
    def load_and_split(self, text_splitter: TextSplitter):
        doc = fitz.open(self.fname)
        for i, page in enumerate(doc): # iterate the document pages
            text = page.get_text()  # In the future, we should get images as well.
            split_text = text_splitter.split_text(text)
            for split in split_text:
                yield RawChunk(split, {'page': i})


class JsonLoader(FileLoader):
    def load_and_split(self, text_splitter: TextSplitter):
        with open(self.fname, 'r') as file:
            original_contents = file.read()
        splitsville = text_splitter.split_text(original_contents)
        for split in splitsville:
            yield RawChunk(split, {})


CODE_SUPPORTS = ['java', 'py', 'cpp', 'c', 'cs', 'js', 'ts', 'rs', 'swift', 'kt', 'm', 'scala', 'lua', 'sh', 'pl', 'sql', 'r']
class CodeLoader(FileLoader):
    def load_and_split(self, text_splitter: TextSplitter):
        all_code = ""
        with open(self.fname, 'r') as fhand:
            all_code = fhand.read()
        # TODO: use special text splitter for code.
        splits = text_splitter.split_text(all_code)
        for split in splits:
            yield RawChunk(page_content=split, metadata={})


class TextLoader(FileLoader):
    def load_and_split(self, text_splitter):
        content = ""
        with open(self.fname, 'r') as file:
            content = file.read()
        splitsville = text_splitter.split_text(content)
        for split in splitsville:
            yield RawChunk(split, {})


class MarkdownLoader(FileLoader):
    def load_and_split(self, text_splitter):
        content = ""
        with open(self.fname, 'r') as file:
            content = file.read()
        splitsville = text_splitter.split_text(content, separators=['\n\n', '#', '##', '###', '\n', '.', ','])
        for split in splitsville:
            yield RawChunk(split, {})


DISALLOWED_DOC_EXTENTIONS = ['pages', 'rtf', 'epub', 'key', 'mp4', 'mov', 'mpeg', 'flv', 'ico']
def get_loader(filetype, path) -> FileLoader:
    if filetype == 'docx':
        return DocxLoader(path)
    if filetype == 'doc':
        return DocLoader(path)
    elif filetype == 'xlsx': 
        return ExcelLoader(path)
    elif filetype in ['html']:
        return HTMLLoader(path)
    elif filetype == 'json':
        return JsonLoader(path)
    elif filetype == 'pdf':
        return PdfLoader(path)
    elif filetype == 'txt':
        return TextLoader(path)
    elif filetype == 'ppt':
        return PptLoader(path)
    elif filetype == 'pptx':
        return PptxLoader(path)
    elif filetype == 'md':
        return MarkdownLoader(path)
    elif filetype == 'odt':
        return OdtLoader(path)
    elif filetype == 'tsv':
        return TsvLoader(path)
    elif filetype == 'csv':
        return CsvLoader(path)
    elif filetype in CODE_SUPPORTS:
        return CodeLoader(path)
    
    # Note on missing file types: EPUB and RTF
    # Both seemed to fail due to missing pandoc installation

    return None



import requests
import sys
from .lm import LM, LMResponse
from pydantic import BaseModel
from .logger import logger


class WebLink():
    def __init__(self, url, name="", language="", snippet="", query="", search_engine="") -> None:
        self.url = url
        self.name = name
        self.language = language
        self.snippet = snippet
        self.query = query
        self.search_engine = search_engine


class SearchEngine():
    def __init__(self, code):
        self.code = code
    
    # Returns tuple (list[WebLink], total: int)
    # The total includes hits not returned (so, it could be like 1 million or something)
    def search(self, query, max_n=10, offset=0):
        raise NotImplementedError("Search not implemented")

    # Returns deduplicated results from many queries
    def search_many(self, queries, max_per=10, offset_for_each=0) -> tuple[list[WebLink], int]:
        all_results = []
        all_urls = {}
        sum_total = 0
        for q in queries:
            results, total = self.search(q, max_n=max_per, offset=offset_for_each)
            results: list[WebLink]
            dedup = [x for x in results if x.url not in all_urls]
            for new_res in dedup:
                all_urls[new_res.url] = True
            all_results.extend(dedup)
            if total is None:
                sum_total = None
            elif sum_total is not None:
                sum_total += total
        return all_results, sum_total

    def gen_queries(self, lm: LM, topic, n=5):

        class Queries(BaseModel):
            queries: list[str]

        intro = f"Your task is to write about {n} search queries such as those that might appear in Google or Bing. Queries might also be put into a special search engine like Reddit or arXiv."
        body = "The queries you write must be in service of a research goal, which will be provided by the user. You should be creative. You might, for example, write queries that could accomplish subgoals within the research topic or lead to surprising new paths. You should use your background knowledge to determine what you might mention in the query."
        formatting = "You must write your queries in the proper JSON schema."
        system = " ".join([intro, body, formatting])

        res: LMResponse = lm.run(topic, system=system, json_schema=Queries)
        parsed: Queries = res.parsed
        return parsed.queries


class Bing(SearchEngine):
    def __init__(self, api_key, market='en-US'):
        self.api_key = api_key
        self.market = market
        super().__init__('bing')

    def search(self, query, max_n=10, offset=0):
        endpoint = "https://api.bing.microsoft.com/v7.0/search"
        # Refer here for bing market codes:  https://learn.microsoft.com/en-us/bing/search-apis/bing-web-search/reference/market-codes
        mkt = self.market
        params = { 'q': query, 'mkt': mkt, 'count': max_n, 'offset': offset }  # count defaults to 10, can be up to 50
        headers = { 'Ocp-Apim-Subscription-Key': self.api_key }
        response = requests.get(endpoint, headers=headers, params=params)
        response.raise_for_status()  # will throw an error if the request isn't good
        my_json = response.json()
        try:
            total = my_json['webPages']['totalEstimatedMatches']
            raw_results = my_json['webPages']['value']
        except:
            logger.error(f"Could not get web pages from search engine. Here was the repsonse: {my_json}")
            return [], 0
        
        results = []
        for i, x in enumerate(raw_results):
            if i >= max_n:
                break
            results.append(WebLink(
                x['url'],
                x['name'],
                language=x['language'],
                snippet=x['snippet'],
                query=query,
                search_engine=self.code
            ))
        return results, total



import logging

DEFAULT_LOG_LEVEL = logging.INFO

logger = logging.getLogger(__name__)
logger.setLevel(DEFAULT_LOG_LEVEL)

handler = logging.StreamHandler()
formatter = logging.Formatter('%(levelname)s - %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

def configure_logging(level=logging.INFO, log_file=None):    
    # Remove existing handlers
    for handler in logger.handlers:
        logger.removeHandler(handler)
    
    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    else:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    logger.setLevel(level)



from urllib.parse import urlparse
import os
from requests.structures import CaseInsensitiveDict
import mimetypes
import tiktoken


def get_mime_type_from_headers(headers: CaseInsensitiveDict):
    content_type: str = headers.get('content-type')
    if not content_type:
        return None
    mime_type = content_type.split(';')[0].strip()
    return mime_type


def get_ext_from_path(path, with_dot=False):
    splitsville = os.path.splitext(path)
    if len(splitsville) > 1:
        has_dot = splitsville[1]
        if not len(has_dot):
            return None
        if with_dot or has_dot[0] != '.':
            return has_dot
        return has_dot[1:]
    return None


# Includes extension
def get_filename_from_path(path):
    splitsville = os.path.split(path)
    if len(splitsville) > 1:
        return splitsville[1]
    return None


def guess_filename_from_url(url):
    parsed_url = urlparse(url)    
    path = parsed_url.path    
    filename = os.path.basename(path)
    if '.' in filename:
        return filename
    else:
        return None


def get_filename_from_headers(headers: CaseInsensitiveDict):
    content_disposition = headers.get('Content-Disposition')

    if content_disposition:
        # Look for the filename in the Content-Disposition header
        filename = None
        cd_parts = content_disposition.split(';')
        for part in cd_parts:
            part: str
            if 'filename=' in part:
                # Strip out the 'filename=' and any surrounding quotes
                filename = part.split('=')[1].strip().strip('"')
                break
        return filename

    return None


# No dot
def get_ext_from_mime_type(mimetype):
    extensions = mimetypes.guess_all_extensions(mimetype)
    if len(extensions):
        with_dot = extensions[0]
        if with_dot[0] == '.':
            return with_dot[1:]
        return with_dot
    return None


def get_mime_type_from_ext(ext):
    # Ensure the extension has a leading dot
    if not ext.startswith('.'):
        ext = '.' + ext
    mime_type, _ = mimetypes.guess_type('file' + ext)
    return mime_type


tokenizer = tiktoken.get_encoding("gpt2")  # GPT2 tokenizer is also used for GPT-3, and maybe GPT-4?
def get_token_estimate(txt):
    tokens = tokenizer.encode(txt)
    return len(tokens)



from .lm import LM, OpenAILM, LMResponse, make_retrieval_prompt
from .search_engine import SearchEngine, WebLink, Bing
from .scraper import ScrapeServ, Scraper, RequestsScraper, ScrapeDataConsumer, ScrapeMetadata, ScrapeResponse, ScrapeScreenshotConsumer
from .wiki import Wiki
from .db import Source, Entity
from .logger import configure_logging
from .exceptions import ContextExceeded, EntityNotExists, SourceNotExists
from .utils import get_token_estimate



from dataclasses import dataclass
import datetime
import sqlite3
from datetime import datetime
import tempfile
import os
from .logger import logger


DB_VERSION = 3  # Incremement when making changes to the schema to force auto migration

ENTITY_TYPES = ["person", "place", "organization", "event", "publication", "law", "product", "object", "concept"]

@dataclass
class Source():
    title: str | None = None
    text: str | None = None

    id: int | None = None
    created_at: datetime | None = None
    are_entities_extracted: int | None = None

    author: str | None = None
    desc: str | None = None
    url: str | None = None
    snippet: str | None = None
    query: str | None = None
    search_engine: str | None = None

@dataclass
class PrimarySourceData():
    source_id: int | None = None
    mimetype: str | None = None
    data: bytes | None = None
    id: int | None = None
    created_at: datetime | None = None

@dataclass
class ScreenshotData():
    source_id: int | None = None
    mimetype: str | None = None
    data: bytes | None = None
    place: int | None = None 
    id: int | None = None
    created_at: datetime | None = None

@dataclass
class Entity():
    slug: str | None = None
    title: str | None = None
    type: str | None = None
    desc: str | None = None
    markdown: str | None = None
    is_written: int | None = None
    id: int | None = None
    created_at: datetime | None = None

@dataclass
class Reference():
    source_id: int = None
    entity_id: int = None
    id: int = None

# Mapping of table names to dataclasses
TABLE_TO_DATACLASS = {
    "sources": Source,
    "primary_source_data": PrimarySourceData,
    "screenshot_data": ScreenshotData,
    "entities": Entity,
    "refs": Reference
}
DATACLASS_TO_TABLE = {v: k for k, v in TABLE_TO_DATACLASS.items()}

def create_db(path):
    conn = sqlite3.connect(path)
    cursor = conn.cursor()
    try:
        sql = """
        CREATE TABLE sources (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            created_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            are_entities_extracted INTEGER DEFAULT 0,
            
            title TEXT,
            text TEXT,

            author TEXT,
            desc TEXT,
            url TEXT,
            snippet TEXT,
            query TEXT,
            search_engine TEXT,

            tbl TEXT DEFAULT 'sources'
        )
        """
        cursor.execute(sql)

        # FTS5 Search over source text
        sql = """
            CREATE VIRTUAL TABLE IF NOT EXISTS sources_fts5 USING fts5(
                source_id UNINDEXED,
                title,
                author,
                desc,
                text
            );
        """
        cursor.execute(sql)
        sql = """
            CREATE TRIGGER IF NOT EXISTS after_sources_insert
            AFTER INSERT ON sources
            BEGIN
                INSERT INTO sources_fts5 (source_id, title, author, desc, text) VALUES (NEW.id, NEW.title, NEW.author, NEW.desc, NEW.text);
            END;
        """
        cursor.execute(sql)
        sql = """
            CREATE TRIGGER IF NOT EXISTS after_sources_update
            AFTER UPDATE ON sources
            BEGIN
                DELETE FROM sources_fts5 WHERE source_id = OLD.id;
                INSERT INTO sources_fts5 (source_id, title, author, desc, text) VALUES (NEW.id, NEW.title, NEW.author, NEW.desc, NEW.text);
            END;
        """
        cursor.execute(sql)
        sql = """
            CREATE TRIGGER IF NOT EXISTS after_sources_delete
            AFTER DELETE ON sources
            BEGIN
                DELETE FROM sources_fts5 WHERE source_id = OLD.id;
            END;
        """
        cursor.execute(sql)

        sql = """
        CREATE TABLE primary_source_data (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            source_id INTEGER,
            created_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            mimetype TEXT,
            data BLOB,
            tbl TEXT DEFAULT 'primary_source_data',
            FOREIGN KEY (source_id) REFERENCES sources(id) ON DELETE CASCADE
        )
        """
        cursor.execute(sql)

        sql = """
            CREATE TABLE screenshot_data (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                source_id INTEGER,
                created_at DATETIME DEFAULT CURRENT_TIMESTAMP,
                mimetype TEXT,
                data BLOB,
                place INTEGER,
                tbl TEXT DEFAULT 'screenshot_data',
                FOREIGN KEY (source_id) REFERENCES sources(id) ON DELETE CASCADE
            )
        """
        cursor.execute(sql)

        sql = """
            CREATE TABLE entities (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                slug VARCHAR(255) UNIQUE,
                created_at DATETIME DEFAULT CURRENT_TIMESTAMP,
                title TEXT,
                type TEXT,
                desc TEXT,
                markdown TEXT,
                is_written INTEGER DEFAULT 0,

                tbl TEXT DEFAULT 'entities'
            )
        """
        cursor.execute(sql)
        sql = """
            CREATE UNIQUE INDEX idx_entities_slug ON entities(slug);
        """
        cursor.execute(sql)

        sql = """
            CREATE TABLE refs (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                source_id INTEGER,
                entity_id INTEGER,
                tbl TEXT DEFAULT 'refs',
                FOREIGN KEY (source_id) REFERENCES sources(id) ON DELETE CASCADE,
                FOREIGN KEY (entity_id) REFERENCES entities(id) ON DELETE CASCADE
            )
        """
        cursor.execute(sql)
        sql = """
            CREATE INDEX idx_references_source_id ON refs(source_id);
        """
        cursor.execute(sql)
        sql = """
            CREATE INDEX idx_references_entity_id ON refs(entity_id);
        """
        cursor.execute(sql)

        sql = f"""
            PRAGMA user_version = {DB_VERSION}
        """
        cursor.execute(sql)
        conn.commit()
    finally:
        conn.close()


def obj_factory(cursor: sqlite3.Cursor, row):
    """Maps database rows to their respective dataclass objects."""
    fields = [column[0] for column in cursor.description]
    my_dict = { key: value for key, value in zip(fields, row) }
    if 'tbl' not in my_dict:  # Would be something special like getting a pragma or something
        return my_dict
    tbl = my_dict['tbl']
    dataclass = TABLE_TO_DATACLASS[tbl]
    del my_dict['tbl']
    return dataclass(**my_dict)


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


# Migrates the user's database to the current version
# Will copy over all columns that still exist + add new ones (with null values)
# Force is just for testing
def migrate_db(path, conn: sqlite3.Connection, force=False):
    sql = "PRAGMA user_version"
    res = conn.execute(sql)
    version = res.fetchone()['user_version']
    if version >= DB_VERSION:
        return path, conn

    logger.warning("Existing wiki database is out of date; migrating to new version.")

    # Create new database
    temp_file = tempfile.NamedTemporaryFile(mode='w+', delete=False)
    db_path = temp_file.name

    create_db(db_path)
    new_conn = sqlite3.connect(db_path)
    new_conn.row_factory = obj_factory

    new_cursor = new_conn.cursor()
    conn.row_factory = dict_factory  # Makes it easier to copy things over, but new one stays obj_factory
    old_cursor = conn.cursor()

    tables_to_reconstruct = [x for x in TABLE_TO_DATACLASS.keys()]

    for table in tables_to_reconstruct:
        # Get common columns first
        new_cursor.execute(f"PRAGMA table_info({table});")
        new_columns = {col["name"] for col in new_cursor.fetchall()}

        old_cursor.execute(f"PRAGMA table_info({table});")
        old_columns = {col["name"] for col in old_cursor.fetchall()}

        common_columns = new_columns.intersection(old_columns)
        if not common_columns:
            continue

        columns_list = ", ".join(common_columns)

        # 1. Read rows from old DB in one go (or chunked if data is huge).
        select_sql = f"SELECT {columns_list} FROM {table}"
        old_cursor.execute(select_sql)
        rows = old_cursor.fetchall()

        # 2. Insert them into new DB efficiently:
        placeholders = ", ".join(["?"] * len(common_columns))
        insert_sql = f"INSERT INTO {table} ({columns_list}) VALUES ({placeholders})"

        # Wrap in a transaction for fewer commits
        new_conn.execute("BEGIN TRANSACTION;")
        # executemany can take an iterable of tuples
        new_cursor.executemany(
            insert_sql,
            (
                tuple(row_dict[col] for col in common_columns)
                for row_dict in rows
            ),
        )
        new_conn.execute("COMMIT;")

    new_conn.commit()
    
    new_cursor.close()
    old_cursor.close()

    # Remove the old database and attach new one
    os.remove(path)
    os.rename(db_path, path)

    new_conn = sqlite3.connect(path)
    new_conn.row_factory = obj_factory

    return path, new_conn



from .db import Source, Entity, PrimarySourceData, ScreenshotData, ENTITY_TYPES, Reference, obj_factory, migrate_db, DATACLASS_TO_TABLE, create_db
from .scraper import Scraper, ScrapeResponse
from .search_engine import WebLink
from .file_loaders import get_loader, FileLoader, RawChunk, TextSplitter
from .utils import get_ext_from_mime_type, get_token_estimate, get_ext_from_path, get_filename_from_path, get_mime_type_from_ext
import os
from pydantic import BaseModel
from .lm import LM, make_retrieval_prompt, LMResponse
from slugify import slugify
import traceback
import sqlite3
from flask import Flask, send_from_directory, jsonify, request
from flask_cors import cross_origin
import re
from .logger import logger
from tqdm import tqdm
import logging
from typing import Match
from .exceptions import EntityNotExists, SourceNotExists


class Wiki():
    title: str
    topic: str
    path: str
    conn: sqlite3.Connection
    def __init__(self, topic: str, title=None, path=None, replace=False):
        self.title = title
        self.topic = topic
        if path is None:
            self.path = 'wiki.sqlite'
        else:
            # TODO: auto populate title / topic, other things
            self.path = path
        
        if replace and os.path.exists(self.path):
            os.remove(self.path)

        if not os.path.exists(self.path):
            create_db(self.path)

        conn: sqlite3.Connection = sqlite3.connect(self.path)
        conn.row_factory = obj_factory
        self.conn = conn
        self._migrate_db()
    

    def _migrate_db(self):
        # Only actually does the migration if necessary
        path, conn = migrate_db(self.path, self.conn)
        self.path = path
        self.conn = conn


    # Just pass the object with the appropriate dataclass, and it will naturally just work
    def _insert_row(self, item):
        dataclass = type(item)
        if dataclass not in DATACLASS_TO_TABLE:
            raise ValueError("Unrecognized dataclass")
        table = DATACLASS_TO_TABLE[dataclass]
        
        fields = [f for f, v in item.__dict__.items() if v is not None]
        placeholders = ", ".join(["?"] * len(fields))
        columns = ", ".join(fields)

        sql = f"""
            INSERT INTO {table} ({columns}) VALUES ({placeholders})
        """
        values = tuple(getattr(item, field) for field in fields)

        cursor: sqlite3.Cursor = self.conn.cursor()
        try:
            cursor.execute(sql, values)
            sql = f"""
                SELECT * FROM {table} WHERE `id`=?
            """
            cursor.execute(sql, (cursor.lastrowid,))
            new_row = cursor.fetchone()
            self.conn.commit()
            return new_row
        finally:
            cursor.close()

    # Updates based on ID (item.id must be set)
    def _update_row(self, item):
        dataclass = type(item)
        if dataclass not in DATACLASS_TO_TABLE:
            raise ValueError("Unrecognized dataclass")
        table = DATACLASS_TO_TABLE[dataclass]
        assert(item.id is not None)

        items = [(k, v) for k, v in item.__dict__.items() if v is not None]  # Extract and force an order
        placeholders = ", ".join([f"{k}=?" for k, _ in items])
        sql = f"""
            UPDATE {table} SET {placeholders} WHERE `id`=?
        """
        values = [v for _, v in items] + [getattr(item, 'id')]

        cursor: sqlite3.Cursor = self.conn.cursor()
        try:
            cursor.execute(sql, values)
            self.conn.commit()
        finally:
            cursor.close()

    # Select all rows with the same column values as item (where item's col value is not None)
    def _match_rows(self, item, limit: int=None, offset: int=None, order_by: list = None):
        dataclass = type(item)
        if dataclass not in DATACLASS_TO_TABLE:
            raise ValueError("Unrecognized dataclass")
        table = DATACLASS_TO_TABLE[dataclass]
        items = [(k, v) for k, v in item.__dict__.items() if v is not None]  # Extract and force an order
        filters = " AND ".join([f"{k}=?" for k, _ in items])
        values = [v for _, v in items]
        sql = f"SELECT * FROM {table}"
        
        if order_by is not None and len(order_by):
            order_by_fields = ", ".join(order_by)
            sql += f" ORDER BY {order_by_fields}"
        if len(filters):
            sql += f" WHERE {filters}"
        if limit is not None:
            sql += "\nLIMIT ?"
            values.append(limit)
        if offset is not None:
            sql += "\nOFFSET ?"
            values.append(offset)

        cursor: sqlite3.Cursor = self.conn.cursor()
        try:
            cursor.execute(sql, values)
            return cursor.fetchall()
        finally:
            cursor.close()
    
    def _match_row(self, item):
        rows = self._match_rows(item)
        if not rows or not len(rows):
            return None
        return rows[0]
    
    def _get_rows_by_ids(self, cls, ids, limit=None, offset=0):
        if cls not in DATACLASS_TO_TABLE:
            raise ValueError("Unrecognized dataclass")
        table = DATACLASS_TO_TABLE[cls]
        cursor: sqlite3.Cursor = self.conn.cursor()
        if not len(ids):
            return []
        try:
            placeholders = ','.join(['?' for _ in ids])
            sql = f"""
            SELECT * FROM {table} WHERE `id` IN ({placeholders})
            """
            cursor.execute(sql, ids)
            return cursor.fetchall()
        finally:
            cursor.close()

    def _delete_matching_rows(self, item):
        dataclass = type(item)
        if dataclass not in DATACLASS_TO_TABLE:
            raise ValueError("Unrecognized dataclass")
        table = DATACLASS_TO_TABLE[dataclass]
        items = [(k, v) for k, v in item.__dict__.items() if v is not None]  # Extract and force an order
        filters = " AND ".join([f"{k}=?" for k, _ in items])
        values = [v for _, v in items]
        sql = f"""
            DELETE FROM {table} WHERE {filters}
        """
        cursor: sqlite3.Cursor = self.conn.cursor()
        try:
            cursor.execute(sql, values)
            self.conn.commit()
        finally:
            cursor.close()

    
    def search_sources_by_text(self, query) -> list[Source]:
        cursor: sqlite3.Cursor = self.conn.cursor()
        try:
            # See specification of an "fts5 string" here: https://www.sqlite.org/fts5.html#full_text_query_syntax
            escaped_query = query.replace('"', '""')
            quoted_query = f'"{escaped_query}"'
            sql = """
                SELECT * FROM sources
                WHERE rowid IN (SELECT source_id FROM sources_fts5 WHERE sources_fts5 MATCH ?)
            """
            cursor.execute(sql, (quoted_query,))
            return cursor.fetchall()
        finally:
            cursor.close()

    def get_referenced_sources(self, entity_id, limit=1000, offset=0) -> list[Source]:
        ref = Reference(entity_id=entity_id)
        refs: list[Reference] = self._match_rows(ref, limit=limit, offset=offset)
        sources = self._get_rows_by_ids(Source, [ref.source_id for ref in refs])
        return sources
    
    def add_reference(self, entity_id, source_id):
        ent = self._match_row(Entity(id=entity_id))
        if not ent:
            raise EntityNotExists(f"Couldn't find entity with ID {entity_id}; make sure you're passing the ID and not the slug!")
        src = self.get_source_by_id(source_id)
        if not src:
            raise SourceNotExists(f"Couldn't find source with id {source_id}")
        ref = Reference(entity_id=entity_id, source_id=source_id)
        new_ref: Reference = self._insert_row(ref)
        return new_ref

    def get_all_sources(self, limit=5000, offset=0) -> list[Source]:
        return self._match_rows(Source(), limit=limit, offset=offset, order_by=['title'])

    def get_all_entities(self, limit=5000, offset=0) -> list[Entity]:
        return self._match_rows(Entity(), limit=limit, offset=offset, order_by=['title'])

    def get_entity_by_slug(self, slug) -> Entity:
        return self._match_row(Entity(slug=slug))
    
    def get_entities_by_type(self, type) -> list[Entity]:
        return self._match_rows(Entity(type=type))
    
    def get_source_by_id(self, id) -> Source:
        return self._match_row(Source(id=id))
    
    def delete_source_by_id(self, id) -> None:
        self._delete_matching_rows(Source(id=id))

    def delete_entity_by_slug(self, slug) -> None:
        self._delete_matching_rows(Entity(slug=slug))

    def add_entity(self, title: str, type: str=None, desc: str=None, markdown: str=None) -> Entity:
        slug = slugify(title)
        is_written = markdown is not None and len(markdown)
        entity = Entity(slug=slug, title=title, type=type, desc=desc, markdown=markdown, is_written=is_written)
        entity = self._insert_row(entity)
        return entity
    
    def update_entity_by_slug(self, slug, title: str = None, type: str = None, desc: str = None, markdown: str=None) -> None:
        existing = self.get_entity_by_slug(slug)
        is_written = existing.is_written or (markdown is not None and len(markdown))
        entity = Entity(
            slug=slug,
            title=title,
            type=type,
            desc=desc,
            markdown=markdown,
            is_written=is_written,
            id=existing.id
        )
        self._update_row(entity)  # updates by id, so need to retrieve ID with the get call first

    def update_source_by_id(self, id, title: str = None, author: str = None, desc: str = None, url: str=None, snippet: str = None, query: str = None) -> None:
        entity = Source(
            id=id,
            title=title,
            desc=desc,
            author=author,
            url=url,
            snippet=snippet,
            query=query
        )
        self._update_row(entity)
    
    def add_source(self, title: str, text: str, author: str=None, desc: str=None, url: str=None, snippet: str=None, query: str=None, search_engine: str=None, are_entities_extracted=False) -> Source:
        source = Source(title=title, text=text, author=author, desc=desc, url=url, snippet=snippet, query=query, search_engine=search_engine, are_entities_extracted=are_entities_extracted)
        source = self._insert_row(source)
        return source

    # TODO: allow breaking up by pages when supported
    def load_local_sources(self, paths: list[str]) -> list[Source]:
        for i in tqdm(range(len(paths)), desc="Loading", total=len(paths)):
            path = paths[i]
            ext = get_ext_from_path(path)
            loader = get_loader(ext, path)
            if not loader:
                logger.warning(f"Couldn't process {path}; skipping.")
            text = ""
            for raw_chunk in loader.load_and_split(TextSplitter()):
                raw_chunk: RawChunk
                text += raw_chunk.page_content
            src: Source = self.add_source(
                title=get_filename_from_path(path),
                text=text
            )

            data = None
            with open(path, 'rb') as fhand:
                data = fhand.read()

            psd = PrimarySourceData(
                source_id=src.id,
                mimetype=get_mime_type_from_ext(ext),
                data=data
            )
            self._insert_row(psd)
        

    # Scrapes and stores info in the db
    def scrape_web_results(self, scraper: Scraper, results: list[WebLink], max_n=None) -> list[tuple[WebLink, Source | None]]:
        scraped = []
        total = min(len(results), max_n) if max_n is not None else len(results)
        for i in tqdm(range(total), desc="Scraping", total=total):
            res = results[i]
            resp: ScrapeResponse = scraper.scrape(res.url)
            new_source = None
            if not resp.success:
                logger.warning(f"Failed to scrape {res.url}; moving on.")
            else:
                with resp.consume_data() as path:
                    with resp.consume_screenshots() as (ss_paths, ss_mimetypes):
                        ext = get_ext_from_mime_type(resp.metadata.content_type)
                        loader: FileLoader = get_loader(ext, path)
                        if not loader:
                            logger.warning(f"Filetype '{resp.metadata.content_type}' cannot be parsed; moving along.")
                        else:
                            txt = ""
                            splitter = TextSplitter()
                            for chunk in loader.load_and_split(splitter):
                                chunk: RawChunk
                                txt += chunk.page_content

                            with open(path, 'rb') as f:
                                file_data = f.read()

                            new_source = Source(
                                title=resp.metadata.title,
                                text=txt,
                                url=resp.url,
                                snippet=res.snippet,
                                query=res.query,
                                search_engine=res.search_engine
                            )
                            new_source: Source = self._insert_row(new_source)

                            primary_data = PrimarySourceData(
                                mimetype=resp.metadata.content_type,
                                data=file_data,
                                source_id=new_source.id,
                            )
                            self._insert_row(primary_data)

                            ss_paths: list[str]
                            ss_mimetypes: list[str]
                            for i, (ss_path, ss_mimetype) in enumerate(zip(ss_paths, ss_mimetypes)):
                                with open(ss_path, 'rb') as f:
                                    ss_data = f.read()
                                
                                screenshot = ScreenshotData(
                                    mimetype=ss_mimetype,
                                    data=ss_data,
                                    source_id=new_source.id,
                                    place=i
                                )
                                self._insert_row(screenshot)
            
            scraped.append((res, new_source))
        return scraped
                    

    def make_entities(self, lm: LM, max_sources=None) -> list[tuple[Source, list[Entity]]]:
        # Go through sources
        sources = self._match_rows(Source(are_entities_extracted=False))
        if not len(sources):
            logger.warning("No sources found to make entities from.")
        
        total = min(len(sources), max_sources) if max_sources is not None else len(sources)
        processed = []
        for i in tqdm(range(total), total=total, desc="Extracting"):
            src: Source = sources[i]
            made_entities = []
            try:
                text = src.text
                if text:
                    not_entities = ["Countries", "Nouns unrelated to the research topic", "Very common nouns or words", "Well-known cities not especially significant to the research topic"]
                    class EntityData(BaseModel):
                        type: str
                        title: str
                        desc: str

                    class Entities(BaseModel):
                        entities: list[EntityData]

                    intro = "You are tasked with extracting relevant entities from the given source material into JSON. Here is the text extracted from the source material:"
                    prompt_src_text = make_retrieval_prompt(lm, [text])
                    wiki = "Each entity will become a custom Wiki article that the user is constructing based on his research topic."
                    type_req = f"The entities/pages can be the following types: {', '.join(ENTITY_TYPES)}"
                    neg_req = f"You should not make entities for the following categories, which don't count: {', '.join(not_entities)}"
                    rel_req = "The user will provide the research topic. THE ENTITIES YOU EXTRACT MUST BE RELEVANT TO THE USER'S RESEARCH GOALS."
                    format_req = "Use the appropriate JSON schema. Here is an example for an extraction for research involving the history of Bell Labs. In this case, we're assuming that the source material mentioned John Bardeen."
                    example = '{"entities": [{"type": "person", "title": "John Bardeen", "desc": "John Bardeen, along with Walter Brattain and Bill Shockley, co-invented the transistor during his time as a physicist at Bell Labs."}, ...]}'
                    example_cont = "For this example, you may also want to have included the transistor (object), The Invention of the Transistor (event), Walter Brattain (person), and Bill Shockley (person), assuming that all of these were actually mentioned in the source material."
                    conclusion = "Now, read the user's research topic and extract the relevant entites from the source given above. Include only entities that were mentioned in the source material. Try to limit the number of entities you extract to only those most relevant (there may be as few as 0 or at most 10)."
                    system_prompt = "\n\n".join([intro, prompt_src_text, wiki, type_req, neg_req, rel_req, format_req, example, example_cont, conclusion])
                    user_prompt = self.topic

                    logger.debug(f"Make Entity User Prompt Length: {get_token_estimate(user_prompt)}")
                    logger.debug(f"Make Entity System Prompt Length: {get_token_estimate(system_prompt)}")

                    resp: LMResponse = lm.run(user_prompt, system_prompt, [], json_schema=Entities)
                    entities: Entities = resp.parsed

                    for ent in entities.entities:
                        ent: EntityData

                        if ent.type not in ENTITY_TYPES:
                            logger.warning(f"Extracted type '{ent.type}' not recognized (title was '{ent.title}', source was '{src.title}')")

                        slug = slugify(ent.title, max_length=255)  # 255 is the max length of the slug text in the database... may want to standardize this somewhere.
                        # Check for duplicate
                        existing = self.get_entity_by_slug(slug)
                        if existing:
                            logger.debug(f"Duplicate entity found for {slug}; not re-adding.")
                            new_reference = Reference(
                                source_id=src.id,
                                entity_id=existing.id
                            )
                            self._insert_row(new_reference)
                        else:
                            new_entity: Entity = self.add_entity(ent.title, type=ent.type, desc=ent.desc)
                            made_entities.append(new_entity)
                            new_reference = Reference(
                                source_id=src.id,
                                entity_id=new_entity.id
                            )
                            self._insert_row(new_reference)
                src.are_entities_extracted = True
                self._update_row(src)
                processed.append((src, made_entities))

            except:
                logger.debug(traceback.format_exc())
                logger.warning(f"An exception occurred trying to parse entities of source with id {src.id}. Moving on.")

        return processed


    def heal_markdown(self, markdown: str) -> str:
        """
        Tries to fix various malformed or inconsistent markdown link patterns to match:
        - [[slug | Entity Title]]
        - [[@ID]] for sources
        Removes or downgrades links if corresponding entity/source doesn't exist.
        """

        # --- Helper functions ---
        def is_numeric(s: str) -> bool:
            return s.isdigit()

        def source_exists(num_str: str) -> bool:
            src = self.get_source_by_id(num_str)
            return src is not None

        def entity_exists(slug_or_title: str) -> bool:
            slug_or_title = slugify(slug_or_title)
            entity = self.get_entity_by_slug(slug_or_title)
            return entity is not None

        # For whatever reason, remove weird bracket lookalikes
        markdown = markdown.replace("【", "[").replace("】", "]")

        # ------------------------------------------------------------------------
        # 1) Transform [text](link) => [[link | text]] or [[@ID]]
        # ------------------------------------------------------------------------
        def replace_square_paren(m: Match) -> str:
            text = m.group(1).strip()
            link = m.group(2).strip()

            # Case: [1](1) => [[@1]]
            if is_numeric(text) and is_numeric(link) and text == link:
                if source_exists(text):
                    return f"[[@{text}]]"
                else:
                    return text

            # Case: [1](foo) => possibly [[@1]] if 1 is a source
            if is_numeric(text):
                if source_exists(text):
                    return f"[[@{text}]]"
                else:
                    return text

            # Otherwise treat link as an entity slug
            if entity_exists(link):
                return f"[[{link} | {text}]]"
            else:
                # Link doesn't exist -> keep just the text
                return text

        pattern_square_paren = re.compile(r'\[([^\]]+)\]\(([^\)]+)\)')
        markdown = pattern_square_paren.sub(replace_square_paren, markdown)
        
        "[[@19]]"
        # ------------------------------------------------------------------------
        # 2) Transform [[text]](link) => [[link | text]] or [[@ID]] 
        # ------------------------------------------------------------------------
        def replace_double_bracket_paren(m: Match) -> str:
            text = m.group(1).strip()
            link = m.group(2).strip()

            # e.g. [[1]](1) => [[@1]]
            if is_numeric(text) and is_numeric(link) and text == link:
                if source_exists(text):
                    return f"[[@{text}]]"
                else:
                    return text

            # Otherwise treat link as an entity slug
            if entity_exists(link):
                return f"[[{link} | {text}]]"
            else:
                return text

        pattern_double_bracket_paren = re.compile(r'\[\[([^\]]+)\]\]\(([^\)]+)\)')
        markdown = pattern_double_bracket_paren.sub(replace_double_bracket_paren, markdown)

        # ------------------------------------------------------------------------
        # 3) Transform double-bracket references [[text]] 
        #    - Single references, or multiple references like [[@1], [@2], [@3]]
        # ------------------------------------------------------------------------

        # (a) Helper to handle a single reference
        def handle_single_reference(txt: str) -> str:
            txt = txt.strip()

            # Case: [[slug | Title]]
            if '|' in txt:
                left, right = [x.strip() for x in txt.split('|', maxsplit=1)]
                if is_numeric(left):
                    if source_exists(left):
                        return f"[[@{left}]]"
                    return right
                if entity_exists(left):
                    return f"[[{left} | {right}]]"
                return right  # fallback

            # Case: numeric => [[@N]] if source
            if is_numeric(txt):
                if source_exists(txt):
                    return f"[[@{txt}]]"
                else:
                    return ""

            # Case: @number => treat as source if it exists
            if len(txt) > 1 and txt[0] == '@':
                num = txt[1:]
                if source_exists(num):
                    return f"[[{txt}]]"
                else:
                    return ""

            # Otherwise treat as entity reference
            if entity_exists(txt):
                return f"[[{txt}]]"
            return txt

        # (b) Detect if the text is a comma-separated list of references (e.g. "@1], [@2]")
        # NEW: We'll use a pattern that matches e.g. `[?@?digits]` repeated with commas.
        # If it doesn't match, we treat it as a single reference.
        pattern_multiple_refs = re.compile(
            r'^\s*(?:\[?\s*@?\d+\s*\]?)(?:\s*,\s*\[?\s*@?\d+\s*\]?)+\s*$'
        )

        # (c) Replace function for the double brackets
        def replace_double_bracket(m: Match) -> str:
            text = m.group(1)  # everything inside [[...]]
            text_stripped = text.strip()

            # If it clearly looks like multiple references, split them up.
            if pattern_multiple_refs.match(text_stripped):
                parts = re.split(r'\s*,\s*', text_stripped)
                results = []
                for part in parts:
                    # remove surrounding brackets from each chunk if present
                    part = part.strip('[] ')
                    results.append(handle_single_reference(part))
                return "".join(results)

            # Otherwise, handle it as a single reference or entity
            return handle_single_reference(text_stripped)

        # Use a pattern that captures everything up to the next "]]"
        pattern_double_bracket = re.compile(r'\[\[(.*?)\]\]')
        markdown = pattern_double_bracket.sub(replace_double_bracket, markdown)

        # ------------------------------------------------------------------------
        # 4) Transform single-bracket sources [@2] 
        # ------------------------------------------------------------------------
        def replace_single_bracket_reference(m: Match) -> str:
            inside = m.group(1).strip()  # e.g. "@2" or "42"
            # If it starts with @, parse the numeric part
            if inside.startswith('@'):
                num_str = inside[1:]
                if is_numeric(num_str) and source_exists(num_str):
                    return f"[[{inside}]]"  # e.g. [[@2]]
                else:
                    return ""  # remove
            else:
                # If it's numeric, try to convert to [@N]
                if is_numeric(inside) and source_exists(inside):
                    return f"[[@{inside}]]"
                else:
                    return f"[{inside}]"  # keep the same

        # Pattern matches [@123] or [123] but not [anything else] and not [[2]]
        # We exclude ! so we don't catch image syntax ![...]
        pattern_single_bracket_ref = re.compile(r'(?<!\!)\[(?!\[)(@?\d+)\](?!\])')
        markdown = pattern_single_bracket_ref.sub(replace_single_bracket_reference, markdown)

        return markdown


    def write_article(self, lm: LM, entity_slug: str) -> Entity | None:

        all_entity_text = ""
        
        ent = self.get_entity_by_slug(entity_slug)
        if not ent:
            raise EntityNotExists(f"Cannot write an article for entity with slug '{entity_slug}' since it doesn't exist.")
        
        all_entities: list[Entity] = self.get_all_entities()
        all_entities_info = [(x.slug, x.title) for x in all_entities]
        all_entity_text = "\n".join([f"[[{x[0]} | {x[1]}]]" for x in all_entities_info])

        matching_sources = self.search_sources_by_text(ent.title)
        direct_sources = self.get_referenced_sources(entity_id=ent.id)

        logger.debug(f"Matching Sources Found: {len(matching_sources)}")
        logger.debug(f"Direct Sources Found: {len(direct_sources)}")

        # Combine the results of both queries, ensuring no duplicates
        matches_ids = {match.id for match in matching_sources}
        combined: list[Source] = matching_sources[:]
        for src in direct_sources:
            if src.id not in matches_ids:
                combined.append(src)

        logger.debug(f"Combined Sources Found: {len(combined)}")

        # Now 'sources' contains all the Source objects associated with the matched SourcePrimaryData
        if not len(combined):
            logger.warning(f"No matching sources found for entity {ent.title}; moving on.")
        else:
            try:
                intro = "You are tasked with writing a full wiki entry for some entity. This wiki is not a general wiki; it is meant to fulfill the research goals set by the user. The user will specify the page you are writing. You **must** write in well-formatted markdown."
                links = "You can specify a link to another wiki entry in Wikilink format like: [[slug | Custom Text]] or [[ Title ]]. Whenever you mention some other entity, you should probably use a link. Here are all the entries in the wiki for your reference:"
                all_entity_text = all_entity_text
                source_instruct = "Below are sources from which you can draw material for your wiki entry. **Use only these sources for the information in your entry. You may not draw from any other outside information you may have.** You'll use the sources' ID numbers to write your citations."
                source_text = make_retrieval_prompt(lm, [data.text for data in combined], prefixes=[f"<START OF SOURCE WITH ID={data.id}>\n{data.title} {'| URL: ' + data.url if data.url else ''}" for data in combined], suffixes=[f"</END OF SOURCE WITH ID={data.id}>" for data in combined])
                references = "In order to cite a source using a footnote, use the syntax '[[@SOURCE_ID]]', with the @ sign. For example, a footnote to source with ID 15 would be [[@15]]. WHENEVER YOU CITE A SOURCE (as opposed to another article) YOU MUST USE THE AT (@) SIGN. **Please include inline references whenever possible!** But do not write them in a separate section."
                
                example_instruct = "Here is an example of what some content might look like in a hypothetical page:"
                example = "## Early Victories\nAmong Napoleon's most important early victories was at the [[Siege of Toulon]], which took place during the [[federalist revolts]]. Napoleon instantly won fame when his plan was credited as being the decisive factor in the battle.[[@14]] He would later parlay his fame into commanding an army to lead an invasion of Italy.[[@19]][[@3]]"

                conclusion = "Now the user will specify the actual wiki page you are tasked with writing."
                system = "\n\n".join([intro, links, all_entity_text, source_instruct, source_text, references, example_instruct, example, conclusion])

                user_prompt_intro = f"You are writing the wiki entry for {ent.title}."
                user_prompt_disambiguation = f"A brief description of this entity for disambiguation: {ent.desc}"
                user_prompt_topic = f"The goal of the wiki is to fulfill this research goal: {self.topic}"
                prompt = "\n".join([user_prompt_intro, user_prompt_disambiguation, user_prompt_topic])
                
                logger.debug(f"Write System Prompt:\n\n{system}")
                logger.debug(f"Write User Prompt:\n\n{prompt}")
                logger.debug(f"Write User Prompt Length: {get_token_estimate(prompt)}")
                logger.debug(f"Write System Prompt Length: {get_token_estimate(system)}")

                lm_resp: LMResponse = lm.run(prompt, system, [])

                new_markdown = lm_resp.text

                logger.debug(f"Raw Markdown:\n\n{new_markdown}")
                # Run a fix of common formatting issues
                new_markdown = self.heal_markdown(new_markdown)
                logger.debug(f"Fixed Markdown:\n\n{new_markdown}")

                ent.markdown = new_markdown
                ent.is_written = True
                self._update_row(ent)
                return ent

            except:
                logger.debug(traceback.format_exc())
                logger.error(f"An exception occurred trying to write entry for {ent.slug}. Moving on.")


    def write_articles(self, lm: LM, max_n=None, rewrite=False) -> list[Entity]:
        all_entities: list[Entity] = self.get_all_entities()
        if rewrite:
            entities = list(all_entities)
        else:
            entities = self._match_rows(Entity(is_written=False))
        written = []
        total = min(len(entities), max_n) if max_n is not None else len(entities)
        for i in tqdm(range(total), desc="Writing", disable=logger.level > logging.INFO):
            ent: Entity = entities[i]
            new_entity = self.write_article(lm, ent.slug)
            if new_entity:
                written.append(new_entity)

        return written


    def deduplicate_entities(self, lm: LM, max_groups=None, group_size=100) -> int:
        all_entities = self.get_all_entities()
        # Bunch them 100 at a time (give or take)
        groups: list[list[Entity]] = []
        i = 0
        while True:
            new_group = all_entities[i:i+group_size]
            if len(new_group) < 3:
                groups[-1].extend(new_group)
                break
            elif len(new_group):
                groups.append(new_group)
            else:
                break
            i += group_size

        total = min(len(groups), max_groups) if max_groups is not None else len(groups)
        n = 0
        for i in tqdm(range(total), desc="Culling", disable=logger.level > logging.INFO):
            this_group = groups[i]
            system = "You are tasked with deduplicating a group of entity titles. Given a list of titles, return a well-formatted JSON object consisting of deduplicated titles. You should combine titles that refer to the same thing or concept. If a title is the plural of another, combine them, leaving only the singular form. You must give a complete list of all the titles that remain. The user will provide the list of titles."
            user = "\n".join([x.title for x in this_group])

            class Deduplicated(BaseModel):
                deduplicated_titles: list[str]

            resp: LMResponse = lm.run(user, system, [], json_schema=Deduplicated)
            parsed: Deduplicated = resp.parsed
            deduped_table = {slugify(k): True for k in parsed.deduplicated_titles}
            for ent in this_group:
                ent: Entity
                if ent.slug not in deduped_table:
                    self.delete_entity_by_slug(ent.slug)
                    n += 1
        return n


    def export(self, dir="output", remove_cross_refs=True, remove_source_refs=True):
        all_written_entities: list[Entity] = self._match_rows(Entity(is_written=True))
        if not os.path.exists(dir):
            os.mkdir(dir)
        for ent in all_written_entities:
            fname = f"{ent.slug}.md"
            with open(os.path.join(dir, fname), 'w') as fhand:
                markdown = ent.markdown

                # Remove references / links if desired
                pattern = r'\[\[\s*(.*?)\s*\]\]'
                def replacer(match):
                    content = match.group(1).strip()
                    if content.startswith('@'):  # Case 1: References like [[ @25 ]]
                        return '' if remove_source_refs else f'[[{content}]]'
                    if '|' in content:  # Case 2: Cross links with [[ | ]] form
                        if remove_cross_refs:
                            # Keep only the right side (alias).
                            parts = content.split('|', 1)
                            return parts[1].strip()
                        return f'[[{content}]]'
                    else:
                        return content if remove_cross_refs else f'[[{content}]]'

                markdown = re.sub(pattern, replacer, markdown)

                fhand.write(markdown)


    def serve(self):
        app = Flask(__name__, static_folder='static')

        # Route to serve static files
        @app.route('/', defaults={'path': ''})
        @app.route('/<path:path>')
        def serve_static_files(path):
            if path != "" and not os.path.splitext(path)[1]:
                path += '.html'
            if path != "" and os.path.exists(app.static_folder + '/' + path):
                return send_from_directory(app.static_folder, path)
            else:
                return send_from_directory(app.static_folder, 'index.html')

        @app.route('/api/index', methods=('GET',))
        @cross_origin()
        def api_index():
            entities = self.get_all_entities()
            data = {
                'entities': [{
                    'title': ent.title,
                    'desc': ent.desc,
                    'slug': ent.slug,
                    'is_written': ent.is_written
                } for ent in entities]
            }
            return jsonify(data)
        
        @app.route('/api/sources', methods=('GET',))
        @cross_origin()
        def api_sources():
            sources = self.get_all_sources()
            data = {
                'sources': [{
                    'title': src.title,
                    'id': src.id,
                } for src in sources]
            }
            return jsonify(data)
        
        @app.route('/api/page', methods=('GET',))
        @cross_origin()
        def api_page():
            entity_slug = request.args.get('slug')
            entity_slug = slugify(entity_slug, max_length=255)
            ent = self.get_entity_by_slug(entity_slug)
            if not ent: 
                return jsonify({'message': f'Entity with slug {entity_slug} not found'}), 404
            data = {
                'entity': {
                    'title': ent.title,
                    'desc': ent.desc,
                    'slug': ent.slug,
                    'markdown': ent.markdown
                }
            }
            return jsonify(data)
        
        @app.route('/api/source', methods=('GET',))
        @cross_origin()
        def api_source():
            source_id = request.args.get('id')
            src = self.get_source_by_id(id=source_id)
            if not src:
                return jsonify({'message': f'Source with id {source_id} not found'}), 404
            data = {
                'source': {
                    'title': src.title,
                    'id': src.id,
                    'url': src.url,
                    'search_engine': src.search_engine,
                    'query': src.query,
                    'snippet': src.snippet
                }
            }
            return jsonify(data)
        
        @app.route('/api/update-entity', methods=('POST',))
        @cross_origin()
        def api_update_entity():
            slug = request.json.get('slug')
            ent = self.get_entity_by_slug(slug=slug)
            if not ent:
                return jsonify({'message': f'Entity with slug {slug} not found'}), 404
            title = request.json.get('title')
            type = request.json.get('type')
            desc = request.json.get('desc')
            markdown = request.json.get('markdown')
            self.update_entity_by_slug(
                slug,
                title=title,
                type=type,
                desc=desc,
                markdown=markdown
            )
            return jsonify({'message': 'success'})
        
        @app.route('/api/delete-entity', methods=('POST',))
        @cross_origin()
        def api_delete_entity():
            slug = request.json.get('slug')
            ent = self.get_entity_by_slug(slug=slug)
            if not ent:
                return jsonify({'message': f'Entity with slug {slug} not found'}), 404
            self.delete_entity_by_slug(slug)
            return jsonify({'message': 'success'})
        
        @app.route('/api/delete-source', methods=('POST',))
        @cross_origin()
        def api_delete_source():
            id = request.json.get('id')
            src = self.get_source_by_id(id=id)
            if not src:
                return jsonify({'message': f'Source with id {id} not found'}), 404
            self.delete_source_by_id(src.id)
            return jsonify({'message': 'success'})
        
        app.run(port=5000, threaded=False)



import tempfile
import os
from urllib.parse import urljoin
from bs4 import BeautifulSoup
from requests_toolbelt.multipart.decoder import MultipartDecoder
from .utils import get_mime_type_from_headers, guess_filename_from_url, get_filename_from_headers
from requests.structures import CaseInsensitiveDict
import requests
import json
import sys
from .logger import logger 


class ScrapeMetadata():
    def __init__(self):
        self.author: str = ""
        self.desc: str = ""
        self.title: str = ""
        self.preview_image_url: str = ""
        self.favicon_url: str = ""
        self.content_type: str = ""


class ScrapeDataConsumer:
    def __init__(self, data_path):
        self.data_path = data_path

    def __enter__(self):
        # Returning the path to the data file
        return self.data_path

    def __exit__(self, exc_type, exc_value, traceback):
        # Remove the temporary file when done
        if os.path.exists(self.data_path):
            os.remove(self.data_path)


class ScrapeScreenshotConsumer:
    def __init__(self, data_paths, data_types):
        self.data_paths = data_paths
        self.data_types = data_types

    def __enter__(self):
        # Returning the path to the data file
        return self.data_paths, self.data_types

    def __exit__(self, exc_type, exc_value, traceback):
        # Remove the temporary file when done
        for path in self.data_paths:
            if os.path.exists(path):
                os.remove(path)


class ScrapeResponse():
    success: bool
    status: int
    url: str
    headers: dict
    data_path: str
    screenshot_paths: list
    screenshot_mimetypes: list
    metadata: ScrapeMetadata
    def __init__(self, success, status, url, headers):
        self.success = success
        self.status = status
        self.url = url
        self.screenshot_paths = []
        self.screenshot_mimetypes = []
        self.headers = CaseInsensitiveDict(headers)
        self._determine_content_type()
        self.data_path = None

    def _determine_content_type(self):
        meta = ScrapeMetadata()
        meta.content_type = get_mime_type_from_headers(self.headers)
        # If it's HTML, we'll get it from the <title> tag, so we don't have to do this here
        if meta.content_type != 'text/html':
            filename = get_filename_from_headers(self.headers)
            if filename:
                meta.title = filename
            else:
                guessed_filename = guess_filename_from_url(self.url)
                if guessed_filename:
                    meta.title = guessed_filename
        self.metadata = meta

    def _set_metadata_from_html(self, content):
        soup = BeautifulSoup(content, 'html.parser')

        # TITLE: Get title if exists
        if soup.title.string:
            self.metadata.title = soup.title.string.strip()

        # DESCRIPTION: Check the open graph description first, then meta description
        og_description = soup.find('meta', attrs={'property': 'og:description', 'content': True})
        if og_description:
            self.metadata.desc = og_description['content'].strip()
        else:
            # Fallback to meta description
            meta_description = soup.find('meta', attrs={'name': 'description', 'content': True})
            if meta_description:
                self.metadata.desc = meta_description['content'].strip()
        
        # AUTHOR: Check Open Graph author first
        og_author = soup.find('meta', attrs={'property': 'og:author', 'content': True})
        if og_author:
            self.metadata.author = og_author['content'].strip()
        else:
            # Fallback to meta author if og:description not found
            meta_author = soup.find('meta', attrs={'name': 'author', 'content': True})
            if meta_author:
                self.metadata.author = meta_author['content'].strip()

        # PREVIEW IMAGE
        og_image = soup.find('meta', attrs={'property': 'og:image', 'content': True})
        if og_image:
            self.metadata.preview_image_url = og_image['content']
        else:
            # Fallback to Twitter image if og:image not found
            twitter_image = soup.find('meta', attrs={'name': 'twitter:image', 'content': True})
            if twitter_image:
                self.metadata.preview_image_url = twitter_image['content']

        # FAVICON
        favicon = soup.find('link', rel=lambda x: x and x.lower() in ['icon', 'shortcut icon'])
        favicon_url = favicon['href'] if favicon and favicon.has_attr('href') else None
        if favicon_url:
            # Resolving the favicon URL against the base URL
            self.metadata.favicon_url = urljoin(self.url, favicon_url)

    def set_data(self, content):
        try:
            tmp = None
            if self.metadata.content_type == 'text/html':
                self._set_metadata_from_html(content)
            tmp = tempfile.NamedTemporaryFile(delete=False)
            tmp.write(content)
            self.data_path = tmp.name
        finally:
            if tmp:
                tmp.close()

    def add_screenshot(self, content, mimetype):
        try:
            tmp = tempfile.NamedTemporaryFile(delete=False)
            tmp.write(content)
            self.screenshot_paths.append(tmp.name)
            self.screenshot_mimetypes.append(mimetype)
        finally:
            tmp.close()

    def consume_data(self):
        # Designed to be used like...
        # with scrape.consume_data() as data_path:
        #     # ... do stuff with data_path
        # And removes the file when done.
        if not self.data_path:
            raise Exception("There is no scraped data")
        curr_data_path = self.data_path
        self.data_path = None
        return ScrapeDataConsumer(curr_data_path)
    
    def consume_screenshots(self):
        if not len(self.screenshot_paths):
            return ScrapeScreenshotConsumer([], [])
        these_paths = list(self.screenshot_paths)
        these_types = list(self.screenshot_mimetypes)
        self.screenshot_paths = []
        self.screenshot_mimetypes = []
        return ScrapeScreenshotConsumer(these_paths, these_types)

    def close(self):
        cleaned = 0
        if self.data_path and os.path.exists(self.data_path):
            os.remove(self.data_path)
            cleaned += 1
        if len(self.screenshot_paths):
            logger.debug(f"Scrape response screnshots being removed in garbage collection; this is a sign of buggy code. Scraped URL was '{self.url}'.")
            for ss in self.screenshot_paths:
                if os.path.exists(ss):
                    os.remove(ss)
                    cleaned += 1
        return cleaned

    def __del__(self):
        # Ensure the temporary file is deleted when the object is destroyed
        try:
            cleaned = self.close()
            if cleaned:
                logger.debug(f"{cleaned} Scrape response data items being removed in garbage collection; this is a sign of buggy code. Scraped URL was '{self.url}'.")

        except Exception as e:
            logger.debug(f"Error cleaning up ScrapeResponse for '{self.url}' during object deletion: {e}")


class Scraper():
    def __init__(self):
        pass

    def scrape(self, url):
        raise NotImplementedError("Scrape is not implemented")
    

class ScrapeServ(Scraper):
    def __init__(self, url="http://localhost:5006", api_key=None):
        self.url = url
        self.api_key = api_key

    def scrape(self, url):
        headers = {
            'Authorization': f'Bearer {self.api_key}'
        }
        data = {
            'url': url
        }
        try:
            response = requests.post(self.url + '/scrape', json=data, headers=headers, timeout=30)
            response.raise_for_status()
            decoder = MultipartDecoder.from_response(response)

            resp = None
            for i, part in enumerate(decoder.parts):
                if i == 0:  # First is some JSON
                    json_part = json.loads(part.content)
                    status = json_part['status']
                    headers = json_part['headers']
                    resp = ScrapeResponse(False, status, url, headers)
                elif i == 1:  # Next is the actual content of the
                    if part.content:
                        resp.success = True
                        resp.set_data(part.content)
                else:  # Other parts are screenshots, if they exist
                    # Annoyingly the headers are bytes not strings for some reason
                    mimetype: bytes = part.headers[b'Content-Type']
                    mimetype: str = mimetype.decode()
                    resp.add_screenshot(part.content, mimetype)
            
            return resp

        except requests.RequestException as e:
            if e.response:
                try:
                    my_json = e.response.json()
                    message = my_json['error']
                    logger.warning(f"Error scraping {url}: {message}")
                except:
                    logger.warning(f"Error scraing {url}, couldn't parse error message ({e})")
            return ScrapeResponse(False, status=None, url=url, headers={})


class RequestsScraper(Scraper):
    def __init__(self):
        pass

    def scrape(self, url):
        try:
            headers = {
                'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3',
                'Accept': 'application/json, text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
                'Accept-Language': 'en-US,en;q=0.9'
            }
            response = requests.get(url, headers=headers, timeout=5)
            response.raise_for_status()  # will throw an error if the request isn't good
            status = response.status_code
            website_data = response.content
            resp_headers = response.headers

            resp = ScrapeResponse(True, status, url, resp_headers)
            resp.set_data(website_data)
            return resp

        except requests.RequestException as e:
            logger.warning(f"Couldn't scrape {url}: {e}")
            if e.response:
                status_code = e.response.status_code
                headers = e.response.headers
                return ScrapeResponse(False, status=status_code, url=url, headers=headers)
            else:
                return ScrapeResponse(False, status=None, url=url, headers={})



from setuptools import setup, find_packages

setup(
    name='nichey',
    version='0.0.3',
    packages=find_packages(exclude=["tests", "tests.*", "user", "user.*"]),
    include_package_data=True,  # Add this line
    package_data={
        'nichey': ['static/*', 'static/**/*'],  # Include static folder and all subdirectories
    },
    install_requires=[
        'requests',
        'openai',
        'pydantic',
        'beautifulsoup4',
        'requests-toolbelt',
        'unstructured[csv,doc,docx,epub,md,msg,odt,org,ppt,pptx,rtf,rst,tsv,xlsx]',
        'pymupdf',
        'tiktoken',
        'python-slugify',
        'flask',
        'flask_cors',
        'tqdm'
    ],
    author='Gordon Kamer',
    author_email='gordon@goodreason.ai',
    description='Build a wiki for your research topic',
    long_description="# Nichey: Generate a wiki for your niche.\n\nFor more information, check out Nichey on [GitHub](https://github.com/goodreasonai/nichey).",
    long_description_content_type='text/markdown',
    url='https://github.com/goodreasonai/nichey',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
)



from openai import OpenAI 
from .utils import get_token_estimate
from .file_loaders import TextSplitter
from .exceptions import ContextExceeded


class LMResponse():
    def __init__(self, text, parsed=None):
        self.text = text
        self.parsed = parsed


class LM():
    def __init__(self, max_input_tokens=8096, accepts_images=False, fail_on_overflow=False):
        self.max_input_tokens = max_input_tokens
        self.accepts_images = accepts_images
        self.fail_on_overflow = fail_on_overflow

    # json_schema is a class that inherits from Pydantic BaseModel (not an object)
    def run(self, prompt, system, context: list, json_schema=None):
        raise NotImplementedError("Run is not implemented")


class OpenAILM(LM):
    def __init__(self, model, max_input_tokens=8096, accepts_images=False, api_key=None, base_url=None, fail_on_overflow=False):
        self.model = model
        self.client = OpenAI(
            base_url=base_url,
            api_key=api_key,
        )
        super().__init__(max_input_tokens, accepts_images, fail_on_overflow)

    def _make_messages(self, prompt, system, context: list):
        messages = [
            {
                'role': 'developer',
                'content': system
            },
            *context,
            {
                'role': 'user',
                'content': prompt,
            }
        ]
        if self.fail_on_overflow:
            total_tokens = sum([get_token_estimate(x['content']) for x in messages])
            if total_tokens > self.max_input_tokens:
                raise ContextExceeded(f"OpenAILM {self.model} was given {total_tokens} but only has room for {self.max_input_tokens}")
        return messages

    def run(self, prompt, system="", context: list = [], json_schema=None):
        messages = self._make_messages(prompt, system, context)
        if json_schema is None:
            chat_completion = self.client.chat.completions.create(
                messages=messages,
                model=self.model,
            )
            return LMResponse(chat_completion.choices[0].message.content)
        else:
            chat_completion = self.client.beta.chat.completions.parse(
                messages=messages,
                model="gpt-4o",
                response_format=json_schema
            )
            msg = chat_completion.choices[0].message
            return LMResponse(msg, parsed=msg.parsed)


"""

Percentage of context length that should be used for retrieval in various places in the code base.
Can / should be changed based on typical model behavior and speed.

Sample outputs:
- 5_000 -> 2_500
- 10_000 -> 7_000
- 32_000 -> 23_500
- 100_000 -> 57_500
- 200_000 -> 107_500

"""
def get_safe_context_length(model: LM):
    # Works like tax brackets
    brackets = [
        (5_000, .5),  # 50% of the first 5000 tokens
        (10_000, .9),  # 90% of tokens 5000-10000
        (32_000, .75),  # etc
        (100_000, .5)
    ]
    cl_remaining = model.max_input_tokens
    safety_cl = 0
    prev = 0
    for bracket in brackets:
        overlap = min(cl_remaining, bracket[0] - prev)
        contribution = overlap * bracket[1]
        safety_cl += contribution
        cl_remaining -= bracket[0] - prev
        prev = bracket[0]
        if cl_remaining <= 0:
            break
    
    if cl_remaining > 0:
        safety_cl += cl_remaining * brackets[-1][1]

    return round(safety_cl)


def make_retrieval_prompt(lm: LM, texts, prefixes: list | str = "<BEGIN SOURCE>", suffixes: list | str = "</END SOURCE>", min_source_length=1000):
    room = get_safe_context_length(lm)

    n_texts = len(texts)

    per_source_token_limit = max(room // n_texts, min_source_length)

    total_tokens_used = 0
    prompt = "\n\n"

    if type(prefixes) == str:
        prefixes = [prefixes for _ in range(len(texts))]
    if type(suffixes) == str:
        suffixes = [suffixes for _ in range(len(texts))]
    
    assert(len(suffixes) == len(texts))
    assert(len(prefixes) == len(prefixes))

    for i, text in enumerate(texts):
        prompt += prefixes[i]
        tokens_used = 0

        prompt_src_text = ""
        splitter = TextSplitter()
        for chunk_txt in splitter.split_text(text):
            new_tokens = get_token_estimate(chunk_txt)
            src_exceeds_limit = new_tokens + tokens_used > per_source_token_limit
            overall_exceeds_limit = new_tokens + tokens_used + total_tokens_used > room
            if src_exceeds_limit or overall_exceeds_limit:
                break
            prompt_src_text += chunk_txt
            tokens_used += new_tokens
        total_tokens_used += tokens_used

        prompt += prompt_src_text
        prompt += f"\n{suffixes[i]}\n\n"

    return prompt
